/* GTK - The GIMP Toolkit
 * Copyright (C) 1995-1997 Peter Mattis, Spencer Kimball and Josh MacDonald
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

 /* simple compass widget was created by Leon Opit (eon.opit@locsoft.com) by modifiying the gtk_dial widget 
    source found in the gtk tutorial.
    l
    I hope it is useful to someone other than me and my boat!
 */
#include <math.h>
#include <stdio.h>
#include <gtk/gtkmain.h>
#include <gtk/gtksignal.h>

#include "gtksimple_compass.h"

#define SCROLL_DELAY_LENGTH  300
#define COMPASS_DEFAULT_SIZE 200

/* Forward declarations */

static void gtk_compass_class_init               (GtkCompassClass    *klass);
static void gtk_compass_init                     (GtkCompass         *compass);
static void gtk_compass_destroy                  (GtkObject        *object);
static void gtk_compass_realize                  (GtkWidget        *widget);
static void gtk_compass_size_request             (GtkWidget      *widget,
					       GtkRequisition *requisition);
static void gtk_compass_size_allocate            (GtkWidget     *widget,
					       GtkAllocation *allocation);
static gint gtk_compass_expose                   (GtkWidget        *widget,
						GdkEventExpose   *event);
static gint gtk_compass_button_press             (GtkWidget        *widget,
						GdkEventButton   *event);
static gint gtk_compass_button_release           (GtkWidget        *widget,
						GdkEventButton   *event);
static gint gtk_compass_motion_notify            (GtkWidget        *widget,
						GdkEventMotion   *event);
static gint gtk_compass_timer                    (GtkCompass         *compass);

static void gtk_compass_update_mouse             (GtkCompass *compass, gint x, gint y);
static void gtk_compass_update                   (GtkCompass *compass);
static void gtk_compass_adjustment_changed       (GtkAdjustment    *adjustment,
						gpointer          data);
static void gtk_compass_adjustment_value_changed (GtkAdjustment    *adjustment,
						gpointer          data);

/* Local data */

static GtkWidgetClass *parent_class = NULL;


GType
gtk_compass_get_type ()
{
  static GType compass_type = 0;

  if (!compass_type)
    {
      static const GTypeInfo compass_info =
      {
	sizeof (GtkCompassClass),
	NULL,
	NULL,
	(GClassInitFunc) gtk_compass_class_init,
	NULL,
	NULL,
	sizeof (GtkCompass),
        0,
	(GInstanceInitFunc) gtk_compass_init,
      };

      compass_type = g_type_register_static (GTK_TYPE_WIDGET, "GtkCompass", &compass_info, 0);
    }

  return compass_type;
}

static void
gtk_compass_class_init (GtkCompassClass *class)
{
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;

  parent_class = gtk_type_class (gtk_widget_get_type ());

  object_class->destroy = gtk_compass_destroy;

  widget_class->realize = gtk_compass_realize;
  widget_class->expose_event = gtk_compass_expose;
  widget_class->size_request = gtk_compass_size_request;
  widget_class->size_allocate = gtk_compass_size_allocate;
  widget_class->button_press_event = gtk_compass_button_press;
  widget_class->button_release_event = gtk_compass_button_release;
  widget_class->motion_notify_event = gtk_compass_motion_notify;
}

static void
gtk_compass_init (GtkCompass *compass)
{
  compass->button = 0;
  compass->policy = GTK_UPDATE_CONTINUOUS;
  compass->timer = 0;
  compass->radius = 0;
  compass->pointer_width = 0;
  compass->angle = 0.0;
  compass->old_value = 0.0;
  compass->old_lower = 0.0;
  compass->old_upper = 0.0;
  compass->adjustment = NULL;
}

GtkWidget*
gtk_compass_new (GtkAdjustment *adjustment)
{
  GtkCompass *compass;

  compass = g_object_new (gtk_compass_get_type (), NULL);

  if (!adjustment)
    adjustment = (GtkAdjustment*) gtk_adjustment_new (0.0, 0.0, 360.0, 1.0, 0.0, 0.0);

  gtk_compass_set_adjustment (compass, adjustment);

  return GTK_WIDGET (compass);
}

static void
gtk_compass_destroy (GtkObject *object)
{
  GtkCompass *compass;

  g_return_if_fail (object != NULL);
  g_return_if_fail (GTK_IS_COMPASS (object));

  compass = GTK_COMPASS (object);

  if (compass->adjustment)
    {
      g_object_unref (GTK_OBJECT (compass->adjustment));
      compass->adjustment = NULL;
    }

  if (GTK_OBJECT_CLASS (parent_class)->destroy)
    (* GTK_OBJECT_CLASS (parent_class)->destroy) (object);
}

GtkAdjustment*
gtk_compass_get_adjustment (GtkCompass *compass)
{
  g_return_val_if_fail (compass != NULL, NULL);
  g_return_val_if_fail (GTK_IS_COMPASS (compass), NULL);

  return compass->adjustment;
}

void
gtk_compass_set_update_policy (GtkCompass      *compass,
			     GtkUpdateType  policy)
{
  g_return_if_fail (compass != NULL);
  g_return_if_fail (GTK_IS_COMPASS (compass));

  compass->policy = policy;
}

void
gtk_compass_set_adjustment (GtkCompass      *compass,
			  GtkAdjustment *adjustment)
{
  g_return_if_fail (compass != NULL);
  g_return_if_fail (GTK_IS_COMPASS (compass));

  if (compass->adjustment)
    {
      g_signal_handlers_disconnect_by_func (GTK_OBJECT (compass->adjustment), NULL, (gpointer) compass);
      g_object_unref (GTK_OBJECT (compass->adjustment));
    }

  compass->adjustment = adjustment;
  g_object_ref (GTK_OBJECT (compass->adjustment));

  g_signal_connect (GTK_OBJECT (adjustment), "changed",
		    GTK_SIGNAL_FUNC (gtk_compass_adjustment_changed),
		    (gpointer) compass);
  g_signal_connect (GTK_OBJECT (adjustment), "value_changed",
		    GTK_SIGNAL_FUNC (gtk_compass_adjustment_value_changed),
		    (gpointer) compass);

  compass->old_value = adjustment->value;
  compass->old_lower = adjustment->lower;
  compass->old_upper = adjustment->upper;

  gtk_compass_update (compass);
}

static void
gtk_compass_realize (GtkWidget *widget)
{
  GtkCompass *compass;
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_COMPASS (widget));

  GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED);
  compass = GTK_COMPASS (widget);

  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.event_mask = gtk_widget_get_events (widget) | 
    GDK_EXPOSURE_MASK | GDK_BUTTON_PRESS_MASK | 
    GDK_BUTTON_RELEASE_MASK | GDK_POINTER_MOTION_MASK |
    GDK_POINTER_MOTION_HINT_MASK;
  attributes.visual = gtk_widget_get_visual (widget);
  attributes.colormap = gtk_widget_get_colormap (widget);

  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;
  widget->window = gdk_window_new (widget->parent->window, &attributes, attributes_mask);

  widget->style = gtk_style_attach (widget->style, widget->window);

  gdk_window_set_user_data (widget->window, widget);

  gtk_style_set_background (widget->style, widget->window, GTK_STATE_ACTIVE);
}

static void 
gtk_compass_size_request (GtkWidget      *widget,
		       GtkRequisition *requisition)
{
  requisition->width = COMPASS_DEFAULT_SIZE;
  requisition->height = COMPASS_DEFAULT_SIZE;
}

static void
gtk_compass_size_allocate (GtkWidget     *widget,
			GtkAllocation *allocation)
{
  GtkCompass *compass;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (GTK_IS_COMPASS (widget));
  g_return_if_fail (allocation != NULL);

  widget->allocation = *allocation;
  compass = GTK_COMPASS (widget);

  if (GTK_WIDGET_REALIZED (widget))
    {

      gdk_window_move_resize (widget->window,
			      allocation->x, allocation->y,
			      allocation->width, allocation->height);

    }
  compass->radius = MIN (allocation->width, allocation->height) * 0.45;
  compass->pointer_width = compass->radius / 5;
}

static gint
gtk_compass_expose (GtkWidget      *widget,
		 GdkEventExpose *event)
{
  GtkCompass *compass;
  GdkPoint points[6];
  gdouble s,c;
  gdouble last, increment;
  GtkStyle      *blankstyle;
  gint xc, yc;
  gint upper, lower;
  gint tick_length;
  gint i, inc;


  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_COMPASS (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  if (event->count > 0)
    return FALSE;
  
  compass = GTK_COMPASS (widget);

/*  gdk_window_clear_area (widget->window,
			 0, 0,
			 widget->allocation.width,
			 widget->allocation.height);
*/
  xc = widget->allocation.width / 2;
  yc = widget->allocation.height / 2;

  upper = compass->adjustment->upper;
  lower = compass->adjustment->lower;

  /* Erase old pointer */

  s = sin (RADIANS(compass->last_angle));
  c = cos (RADIANS(compass->last_angle));
  compass->last_angle = compass->angle;

  points[0].x = xc + c*compass->pointer_width/2;
  points[0].y = yc + s*compass->pointer_width/2;
  points[1].x = xc + s*compass->radius;
  points[1].y = yc - c*compass->radius;
  points[2].x = xc - c*compass->pointer_width/2;
  points[2].y = yc - s*compass->pointer_width/2;
  points[3].x = xc - s*compass->radius/1.2;
  points[3].y = yc + c*compass->radius/1.2;
  points[4].x = points[0].x;
  points[4].y = points[0].y;

  blankstyle = gtk_style_new ();
  blankstyle->bg_gc[GTK_STATE_NORMAL] =
                widget->style->bg_gc[GTK_STATE_NORMAL];
  blankstyle->dark_gc[GTK_STATE_NORMAL] =
                widget->style->bg_gc[GTK_STATE_NORMAL];
  blankstyle->light_gc[GTK_STATE_NORMAL] =
                widget->style->bg_gc[GTK_STATE_NORMAL];
  blankstyle->black_gc =
                widget->style->bg_gc[GTK_STATE_NORMAL];

  gtk_paint_polygon (blankstyle,
                    widget->window,
                    GTK_STATE_NORMAL,
                    GTK_SHADOW_OUT,
	            NULL,
                    widget,
                    NULL,
                    points, 5,
                    FALSE);

  g_object_unref (blankstyle);


  /* Draw ticks */

  if ((upper - lower) == 0)
    return FALSE;

      for (i=0; i<360; i++) {
          if (i == 0) {
              tick_length = compass->pointer_width * 5;
          }
          else if (i % 90 == 0) tick_length = compass->pointer_width * 4;
          else if (i % 45 == 0) tick_length = compass->pointer_width * 3;
          else if (i % 30 == 0)tick_length = compass->pointer_width * 2;
          else if (i % 10 == 0)tick_length = compass->pointer_width;
          else  tick_length = compass->pointer_width /2;
    
          s = sin (RADIANS(i));
          c = cos (RADIANS(i));
    
          gdk_draw_line (widget->window,
                         widget->style->fg_gc[widget->state],
                         xc + s*(compass->radius - tick_length),
                         yc - c*(compass->radius - tick_length),
                         xc + s*compass->radius,
                         yc - c*compass->radius);
        }
  /* Draw pointer */

  s = sin (RADIANS(compass->angle));
  c = cos (RADIANS(compass->angle));
  compass->last_angle = compass->angle;

  points[0].x = xc + c*compass->pointer_width/2;
  points[0].y = yc + s*compass->pointer_width/2;
  points[1].x = xc + s*compass->radius;
  points[1].y = yc - c*compass->radius;
  points[2].x = xc - c*compass->pointer_width/2;
  points[2].y = yc - s*compass->pointer_width/2;
  points[3].x = xc - s*compass->radius/1.2;
  points[3].y = yc + c*compass->radius/1.2;
  points[4].x = points[0].x;
  points[4].y = points[0].y;


  gtk_paint_polygon (widget->style,
		    widget->window,
		    GTK_STATE_NORMAL,
		    GTK_SHADOW_OUT,
	            NULL,
                    widget,
                    NULL,
		    points, 5,
		    TRUE);


  return FALSE;
}

static gint
gtk_compass_button_press (GtkWidget      *widget,
		       GdkEventButton *event)
{
  GtkCompass *compass;
  gint dx, dy;
  double s, c;
  double d_parallel;
  double d_perpendicular;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_COMPASS (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  compass = GTK_COMPASS (widget);

  /* Determine if button press was within pointer region - we 
     do this by computing the parallel and perpendicular distance of
     the point where the mouse was pressed from the line passing through
     the pointer */
  
  dx = event->x - widget->allocation.width / 2;
  dy = widget->allocation.height / 2 - event->y;
  
  s = sin (RADIANS(compass->angle));
  c = cos (RADIANS(compass->angle));
  
  d_parallel = c*dy + s*dx;
  d_perpendicular = fabs (c*dx - s*dy);
  
  if (!compass->button &&
      (d_perpendicular < compass->pointer_width/2) &&
      (d_parallel > - compass->pointer_width))
    {
      gtk_grab_add (widget);

      compass->button = event->button;

      gtk_compass_update_mouse (compass, event->x, event->y);
    }

  return FALSE;
}

static gint
gtk_compass_button_release (GtkWidget      *widget,
			  GdkEventButton *event)
{
  GtkCompass *compass;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_COMPASS (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  compass = GTK_COMPASS (widget);

  if (compass->button == event->button)
    {
      gtk_grab_remove (widget);

      compass->button = 0;

      if (compass->policy == GTK_UPDATE_DELAYED)
	gtk_timeout_remove (compass->timer);
      
      if ((compass->policy != GTK_UPDATE_CONTINUOUS) &&
	  (compass->old_value != compass->adjustment->value))
	g_signal_emit_by_name (GTK_OBJECT (compass->adjustment), "value_changed");
    }

  return FALSE;
}

static gint
gtk_compass_motion_notify (GtkWidget      *widget,
			 GdkEventMotion *event)
{
  GtkCompass *compass;
  GdkModifierType mods;
  gint x, y, mask;

  g_return_val_if_fail (widget != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_COMPASS (widget), FALSE);
  g_return_val_if_fail (event != NULL, FALSE);

  compass = GTK_COMPASS (widget);

  if (compass->button != 0)
    {
      x = event->x;
      y = event->y;

      if (event->is_hint || (event->window != widget->window))
	gdk_window_get_pointer (widget->window, &x, &y, &mods);

      switch (compass->button)
	{
	case 1:
	  mask = GDK_BUTTON1_MASK;
	  break;
	case 2:
	  mask = GDK_BUTTON2_MASK;
	  break;
	case 3:
	  mask = GDK_BUTTON3_MASK;
	  break;
	default:
	  mask = 0;
	  break;
	}

      if (mods & mask)
	gtk_compass_update_mouse (compass, x,y);
    }

  return FALSE;
}

static gint
gtk_compass_timer (GtkCompass *compass)
{
  g_return_val_if_fail (compass != NULL, FALSE);
  g_return_val_if_fail (GTK_IS_COMPASS (compass), FALSE);

  if (compass->policy == GTK_UPDATE_DELAYED)
    g_signal_emit_by_name (GTK_OBJECT (compass->adjustment), "value_changed");

  return FALSE;
}

static void
gtk_compass_update_mouse (GtkCompass *compass, gint x, gint y)
{
  gint xc, yc;
  gfloat old_value;

  g_return_if_fail (compass != NULL);
  g_return_if_fail (GTK_IS_COMPASS (compass));

  xc = GTK_WIDGET(compass)->allocation.width / 2;
  yc = GTK_WIDGET(compass)->allocation.height / 2;

  old_value = compass->adjustment->value;
  compass->angle = round(DEGREES (atan2( xc-x,y-yc))); 

  compass->adjustment->value = compass->angle;
  if (compass->adjustment->value != old_value)
    {
      if (compass->policy == GTK_UPDATE_CONTINUOUS)
	{
	  g_signal_emit_by_name (GTK_OBJECT (compass->adjustment), "value_changed");
	}
      else
	{
	  gtk_widget_queue_draw (GTK_WIDGET (compass));

	  if (compass->policy == GTK_UPDATE_DELAYED)
	    {
	      if (compass->timer)
		gtk_timeout_remove (compass->timer);

	      compass->timer = gtk_timeout_add (SCROLL_DELAY_LENGTH,
					     (GtkFunction) gtk_compass_timer,
					     (gpointer) compass);
	    }
	}
    }
}

static void
gtk_compass_update (GtkCompass *compass)
{
  gfloat new_value;
  
  g_return_if_fail (compass != NULL);
  g_return_if_fail (GTK_IS_COMPASS (compass));

  new_value = compass->adjustment->value;
  
  if (new_value < compass->adjustment->lower)
    new_value = compass->adjustment->upper;

  if (new_value > compass->adjustment->upper)
    new_value = compass->adjustment->lower;

  if (new_value != compass->adjustment->value)
    {
      compass->adjustment->value = new_value;
      g_signal_emit_by_name (GTK_OBJECT (compass->adjustment), "value_changed");
    }

  compass->angle = compass->adjustment->value;

  gtk_widget_queue_draw (GTK_WIDGET (compass));
}

static void
gtk_compass_adjustment_changed (GtkAdjustment *adjustment,
			      gpointer       data)
{
  GtkCompass *compass;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);

  compass = GTK_COMPASS (data);

  if ((compass->old_value != adjustment->value) ||
      (compass->old_lower != adjustment->lower) ||
      (compass->old_upper != adjustment->upper))
    {
      gtk_compass_update (compass);

      compass->old_value = adjustment->value;
      compass->old_lower = adjustment->lower;
      compass->old_upper = adjustment->upper;
    }
}

static void
gtk_compass_adjustment_value_changed (GtkAdjustment *adjustment,
				    gpointer       data)
{
  GtkCompass *compass;

  g_return_if_fail (adjustment != NULL);
  g_return_if_fail (data != NULL);

  compass = GTK_COMPASS (data);

  if (compass->old_value != adjustment->value)
    {
      gtk_compass_update (compass);

      compass->old_value = adjustment->value;
    }
}


